use actix_web::error::ErrorInternalServerError;
use image::{imageops::FilterType, DynamicImage, ImageFormat};
use log::{warn, error};
use std::{collections::HashMap, io::Cursor};

use crate::{app_state::AppState, s3_utils::upload_to_s3};

pub const THUMB_WIDTHS: [u32; 7] = [10, 40, 110, 300, 600, 800, 1400];

/// Парсит путь к файлу, извлекая оригинальное имя, требуемую ширину и формат.
/// Примеры:
/// - "filename_150.ext" -> ("filename", 150, "ext")
/// - "unsafe/1440x/production/image/439efaa0-816f-11ef-b201-439da98539bc.jpg" -> ("439efaa0-816f-11ef-b201-439da98539bc", 1440, "jpg")
/// - "unsafe/production/image/5627e002-0c53-11ee-9565-0242ac110006.png" -> ("5627e002-0c53-11ee-9565-0242ac110006", 0, "png")
/// - "unsafe/development/image/439efaa0-816f-11ef-b201-439da98539bc.jpg/webp" -> ("439efaa0-816f-11ef-b201-439da98539bc", 0, "webp")
/// - "059ab1e0-66c7-11e9-adda-e3ab2ab4e37e_800.jpg" -> ("059ab1e0-66c7-11e9-adda-e3ab2ab4e37e", 800, "jpg")
pub fn parse_file_path(requested_path: &str) -> (String, u32, String) {
    let mut path = requested_path.to_string();
    // Удаляем /webp в конце, если есть
    if path.ends_with("/webp") {
        path = path.replace("/webp", "");
    }

    // Получаем последнюю часть пути (имя файла с расширением)
    let filename = path.split('/').last().unwrap_or(&path);
    
    // Разделяем имя файла и расширение
    let (mut base_filename, extension) = filename
        .rsplit_once('.')
        .unwrap_or((filename, ""));
    
    // Инициализируем переменные
    let mut width = 0;
    
    // Проверяем наличие ширины в имени файла
    if let Some((name, width_str)) = base_filename.rsplit_once('_') {
        if let Ok(w) = width_str.parse::<u32>() {
            width = w;
            base_filename = name;
        }
    }
    
    // Если ширина не найдена и путь начинается с "unsafe",
    // ищем ширину в формате "1440x"
    if width == 0 && path.starts_with("unsafe") {
        for part in path.split('/') {
            if part.ends_with('x') {
                if let Ok(w) = part.trim_end_matches('x').parse::<u32>() {
                    width = w;
                    break;
                }
            }
        }
    }

    (base_filename.to_string(), width, extension.to_string())
}

/// Генерирует миниатюры изображения.
///
/// Теперь функция принимает дополнительный параметр `format`, который определяет формат сохранения миниатюр.
/// Это позволяет поддерживать различные форматы изображений без необходимости заранее предугадывать их.
pub async fn generate_thumbnails(
    image: &DynamicImage,
    format: ImageFormat
) -> Result<HashMap<u32, Vec<u8>>, actix_web::Error> {
    let mut thumbnails = HashMap::new();

    for &width in THUMB_WIDTHS.iter().filter(|&&w| w < image.width()) {
        let thumbnail = image.resize(width, u32::MAX, FilterType::Lanczos3); // Ресайз изображения по ширине
        let mut buffer = Vec::new();
        thumbnail
            .write_to(&mut Cursor::new(&mut buffer), format)
            .map_err(|e| {
                log::error!("Ошибка при сохранении миниатюры: {}", e);
                ErrorInternalServerError("Не удалось сгенерировать миниатюру")
            })?; // Сохранение изображения в указанном формате
        thumbnails.insert(width, buffer);
    }

    Ok(thumbnails)
}

/// Определяет формат изображения на основе расширения файла.
fn determine_image_format(extension: &str) -> Result<ImageFormat, actix_web::Error> {
    match extension.to_lowercase().as_str() {
        "jpg" | "jpeg" => Ok(ImageFormat::Jpeg),
        "png" => Ok(ImageFormat::Png),
        "gif" => Ok(ImageFormat::Gif),
        "bmp" => Ok(ImageFormat::Bmp),
        "ico" => Ok(ImageFormat::Ico),
        "tiff" => Ok(ImageFormat::Tiff),
        "webp" => Ok(ImageFormat::WebP),
        _ => {
            log::error!("Неподдерживаемый формат изображения: {}", extension);
            Err(ErrorInternalServerError("Неподдерживаемый формат изображения"))
        },
    }
}

/// Сохраняет данные миниатюры.
/// 
/// Обновлена для передачи корректного формата изображения.
pub async fn thumbdata_save(
    original_data: Vec<u8>,
    state: &AppState,
    original_filename: &str,
    content_type: String,
) -> Result<(), actix_web::Error> {
    if content_type.starts_with("image") {
        warn!("[THUMB] Processing image: {}", original_filename);
        let (base_filename, _, extension) = parse_file_path(&original_filename);
        warn!("[THUMB] Parsed: base={}, ext={}", base_filename, extension);
        
        let ext = extension.to_lowercase();
        let filename = format!("{}.{}", base_filename, ext);
        warn!("[THUMB] Normalized filename: {}", filename);

        let img = match image::load_from_memory(&original_data) {
            Ok(img) => {
                warn!("[THUMB] Successfully loaded image from memory, size: {}x{}", 
                    img.width(), img.height());
                img
            },
            Err(e) => {
                error!("[THUMB] Failed to load image from memory: {}", e);
                return Err(ErrorInternalServerError("cant load image"));
            }
        };

        let format = match determine_image_format(&ext) {
            Ok(f) => {
                warn!("[THUMB] Using format: {:?}", f);
                f
            },
            Err(e) => {
                error!("[THUMB] Format error: {}", e);
                return Err(e);
            }
        };

        match generate_thumbnails(&img, format).await {
            Ok(thumbnails_bytes) => {
                warn!("[THUMB] Generated {} thumbnails", thumbnails_bytes.len());
                
                for (thumb_width, thumbnail) in thumbnails_bytes {
                    let thumb_filename = format!("{}_{}.{}", base_filename, thumb_width, ext);
                    warn!("[THUMB] Saving thumbnail: {}", thumb_filename);
                    
                    match upload_to_s3(
                        &state.storj_client,
                        &state.bucket,
                        &thumb_filename,
                        thumbnail.clone(),
                        &content_type,
                    ).await {
                        Ok(_) => {
                            warn!("[THUMB] Successfully saved thumbnail: {}", thumb_filename);
                            // Сохраняем путь к миниатюре в Redis
                            state.set_path(&thumb_filename, &thumb_filename).await;
                            warn!("[THUMB] Cached path in Redis: {}", thumb_filename);
                        },
                        Err(e) => {
                            error!("[THUMB] Failed to save thumbnail {}: {}", thumb_filename, e);
                        }
                    }
                }
                Ok(())
            },
            Err(e) => {
                error!("[THUMB] Failed to generate thumbnails: {}", e);
                Err(e)
            }
        }
    } else {
        warn!("[THUMB] Skipping non-image content type: {}", content_type);
        Ok(())
    }
}

/// Выбирает ближайший подходящий размер из предопределённых.
/// Если `requested_width` больше максимальной ширины в `THUMB_WIDTHS`,
/// возвращает максимальную ширину.
pub fn find_closest_width(requested_width: u32) -> u32 {
    // Проверяем, превышает ли запрошенная ширина максимальную доступную ширину
    if requested_width > *THUMB_WIDTHS.last().unwrap() {
        return *THUMB_WIDTHS.last().unwrap();
    }

    // Находим ширину с минимальной абсолютной разницей с запрошенной
    *THUMB_WIDTHS
        .iter()
        .min_by_key(|&&width| (width as i32 - requested_width as i32).abs())
        .unwrap_or(&THUMB_WIDTHS[0]) // Возвращаем самый маленький размер, если ничего не подошло
}
